import pyjq
import re

from fimutil.ralph.ralph_uri import RalphURI
from fimutil.ralph.asset import RalphAsset, RalphAssetType


class SimpleModel(RalphAsset):
    """
    Some assets have a separate model entry that behaves like an asset.
    """
    FIELD_MAP = '{Model: .category.name}'

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)
        self.type = RalphAssetType.Model

    def parse(self):
        super().parse()


class WorkerModel(SimpleModel):
    """
    Worker model has extra info like total RAM and cores that we need
    """
    FIELD_MAP = '{Model: .category.name, RAM: .custom_fields.total_memory_ram, ' \
                'CPU: .custom_fields.cpu_socket_count, Core: .cores_count}'
    # .*? to make the first match lazy
    DISK_REGEX = ".*?([\\d.]+)([TGM]B).*"

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)

    def parse(self):
        super().parse()
        # some workers have sas disks - need to count them and sum up disk space
        sas_disk_count = 0
        sas_disk_size = 0.0
        unit = 'TB'
        try:
            sas_disk_count_str = pyjq.one('.custom_fields.sas_disk_count', self.raw_json_obj)
            if sas_disk_count_str is not None:
                sas_disk_count = int(sas_disk_count_str)
            sas_disk_desc = pyjq.one('.custom_fields.sas_disk', self.raw_json_obj)
            if sas_disk_desc is not None:
                match = re.match(self.DISK_REGEX, sas_disk_desc)
                if match is not None:
                    sas_disk_size = float(match.group(1))
                    unit = match.group(2)
        except ValueError:
            pass
        total_disk = sas_disk_size * sas_disk_count
        self.fields['Disk'] = str(total_disk) + ' ' + unit


class StorageModel(SimpleModel):
    """
    Storage model has extra info like total disks
    """
    FIELD_MAP = '{Model: .category.name}'
    # .*? to make the first match lazy
    DISK_REGEX = ".*?([\\d.]+)([TGM]B).*"

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)

    def parse(self):
        super().parse()
        # storage has SAS disks
        sas_disk_count = 0
        sas_disk_size = 0.0
        unit = 'TB'
        try:
            sas_disk_count_str = pyjq.one('.custom_fields.sas_disk_count', self.raw_json_obj)
            if sas_disk_count_str is not None:
                sas_disk_count = int(sas_disk_count_str)
            sas_disk_desc = pyjq.one('.custom_fields.sas_disk', self.raw_json_obj)
            if sas_disk_desc is not None:
                match = re.match(self.DISK_REGEX, sas_disk_desc)
                if match is not None:
                    sas_disk_size = float(match.group(1))
                    unit = match.group(2)
        except ValueError:
            pass
        total_disk = sas_disk_size * sas_disk_count
        self.fields['Disk'] = str(total_disk) + ' ' + unit



import pyjq

from fimutil.ralph.ralph_uri import RalphURI
from fimutil.ralph.asset import RalphAsset, RalphAssetType, RalphAssetMimatch
from fimutil.ralph.model import StorageModel


class Storage(RalphAsset):
    """
    Storage array has pretty minimal information
    """
    FIELD_MAP = '{Name: .hostname, SN: .sn}'

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)
        self.type = RalphAssetType.Storage
        self.model = None

    def parse(self):
        super().parse()

        # find model
        model_url = pyjq.one('.model.url', self.raw_json_obj)
        self.model = StorageModel(uri=model_url, ralph=self.ralph)
        try:
            self.model.parse()
        except RalphAssetMimatch:
            pass

    def __str__(self):
        ret = super().__str__()
        return ret + '\n\t' + str(self.model)



import logging
from abc import ABC
from urllib.parse import urlencode
import pyjq

from fimutil.ralph.ralph_uri import RalphURI
from fimutil.ralph.worker_node import WorkerNode
from fimutil.ralph.storage import Storage
from fimutil.ralph.dp_switch import DPSwitch


class Site:
    """
    As site consists of some number of assets - for information model purposes
    typically some number of worker nodes, a storage node and a dataplane switch.
    """
    def __init__(self, *, site_name: str, ralph: RalphURI, domain: str='.fabric-testbed.net'):
        """
        Site name can be upper or lower case
        """
        self.workers = list()
        self.storage = None
        self.dp_switch = None
        self.name = site_name
        self.domain = domain
        self.ralph = ralph

    def catalog(self):
        """
        Catalog what the site has by querying for regular expressions. Names of nodes
        are expected to conform to the following convention:
        - workers: <site>-w[1234567890]+.fabric-testbed.net
        - storage: <site>-storage.fabric-testbed.net
        - dp switch: <site>-data-sw.fabric-testbed.net
        """
        query = {'hostname__regex': f'{self.name.lower()}-w[1234567890]+' + self.domain}

        results = self.ralph.get_json_object(self.ralph.base_uri + 'data-center-assets/?' +
                                             urlencode(query))
        worker_urls = pyjq.one('[ .results[].url ]', results)

        for worker in worker_urls:
            ralph_worker = WorkerNode(uri=worker, ralph=self.ralph)
            ralph_worker.parse()
            self.workers.append(ralph_worker)

        query = {'hostname': f'{self.name.lower()}-storage' + self.domain}
        results = self.ralph.get_json_object(self.ralph.base_uri + 'data-center-assets/?' +
                                             urlencode(query))
        storage_url = None
        try:
            storage_url = pyjq.one('[ .results[0].url ]', results)[0]
        except ValueError:
            logging.warning('Unable to find storage node in site, continuing')
        self.storage = Storage(uri=storage_url, ralph=self.ralph)
        self.storage.parse()

        query = {'hostname': f'{self.name.lower()}-data-sw' + self.domain}
        results = self.ralph.get_json_object(self.ralph.base_uri + 'data-center-assets/?' +
                                             urlencode(query))
        dp_switch_url = None
        try:
            dp_switch_url = pyjq.one('[ .results[0].url ]', results)[0]
        except ValueError:
            logging.warning('Unable to find a dataplane switch in site, continuing')
        self.dp_switch = DPSwitch(uri=dp_switch_url, ralph=self.ralph)
        self.dp_switch.parse()

    def __str__(self):
        assets = list()
        if self.storage is not None:
            assets.append(str(self.storage))
        if self.dp_switch is not None:
            assets.append(str(self.dp_switch))
        for w in self.workers:
            assets.append(str(w))
        return '\n'.join(assets)



"""
Scan workers given on command line printing to stdout all pertinent info
"""

import argparse
import traceback
import logging
import sys

from fimutil.ralph.ralph_uri import RalphURI
from fimutil.ralph.worker_node import WorkerNode

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    # split into different mutually exclusive operations

    parser.add_argument("-w", "--worker", action="store",
                        help="Scan single worker information")
    parser.add_argument("-b", "--base_uri", action="store",
                        help="Base URL of API")
    parser.add_argument("-d", "--debug", action="count",
                        help="Turn on debugging")
    parser.add_argument("-t", "--token", action="store",
                        help="Ralph API token value")

    args = parser.parse_args()

    if args.debug is None:
        logging.basicConfig(level=logging.INFO)
    elif args.debug >= 1:
        logging.basicConfig(level=logging.DEBUG)

    if args.worker is None:
        print('You must specify the worker', file=sys.stderr)
        sys.exit(-1)

    if args.base_uri is None:
        print('You must specify the base URL (typically https://hostname/api/data-center-assets/',
              file=sys.stderr)
        sys.exit(-1)

    if args.token is None:
        print('You must specify a Ralph API token, you can find it in your profile page in Ralph',
              file=sys.stderr)
        sys.exit(-1)

    ralph = RalphURI(token=args.token, base_uri=args.base_uri)
    worker = WorkerNode(uri=args.base_uri + 'data-center-assets/?hostname=' + args.worker,
                        ralph=ralph)
    worker.parse()
    print(worker)




from typing import List, Any
from dataclasses import dataclass

import pyjq

GPU_MODELS = ['Quadro RTX 6000/8000', 'Tesla T4']


@dataclass(frozen=True)
class GPU:
    """
    Note that because GPUs aren't properly part of Ralph catalog structure,
    they appear as special fields inside the node and as such are not
    subclassed from RalphAsset (they don't have a URL)
    """
    Model: str
    Description: str
    BDF: str

    @staticmethod
    def find_gpus(node_raw_json) -> List[Any]:
        """
        Find if there are GPUs in this node. Returns a list
        of GPU objects (which can be empty).
        """
        custom_fields = pyjq.one('.custom_fields', node_raw_json)
        ret = list()
        for field in custom_fields:
            for gpu_model in GPU_MODELS:
                if gpu_model in custom_fields[field]:
                    model = gpu_model
                    description = custom_fields[field]
                    bdf = custom_fields[field + '_pci_id']
                    ret.append(GPU(model, description, bdf))
        return ret


import urllib3
import json


class RalphURI:
    """
    Load JSON file from a Ralph URI. Deal with authentication.
    """

    def __init__(self, *, token: str, base_uri: str):
        self.token = token
        self.base_uri = base_uri
        self.pool = urllib3.PoolManager()

    def get_json_object(self, uri: str):
        # avoid http->https redirects, just replace directly in uri
        uri = uri.replace('http:', 'https:')
        if not uri.startswith(self.base_uri):
            raise RalphURIError(msg=f'Provided uri {uri} does not match base uri {self.base_uri}')
        headers = {'Authorization': f'Token {self.token}', 'Content-Type': 'application/json'}
        r = self.pool.request("GET", uri, headers=headers)
        if r.status != 200:
            raise RuntimeError(f'Unable to contact {uri=} due to error {r.status=}')

        return json.loads(r.data.decode("utf-8"))


class RalphURIError(Exception):
    def __init__(self, msg: str):
        super().__init__(f"Ralph URI Error: {msg}")



import pyjq
import logging

from fimutil.ralph.ralph_uri import RalphURI
from fimutil.ralph.asset import RalphAsset, RalphAssetType, RalphAssetMimatch
from fimutil.ralph.model import SimpleModel
from fimutil.ralph.ethernetport import EthernetPort


class DPSwitch(RalphAsset):
    """
    Dataplane switch
    """
    FIELD_MAP = '{Name: .hostname, SN: .sn}'

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)
        self.type = RalphAssetType.DPSwitch
        self.model = None

    def parse(self):
        super().parse()

        # find model
        model_url = pyjq.one('.model.url', self.raw_json_obj)
        self.model = SimpleModel(uri=model_url, ralph=self.ralph)
        try:
            self.model.parse()
        except RalphAssetMimatch:
            logging.warning('Unable to parse switch model, continuing')

        try:
            port_urls = pyjq.all('.ethernet[].url', self.raw_json_obj)
        except ValueError:
            logging.warning('Unable to find any ethernet ports in node, continuing')
            port_urls = list()

        port_index = 1
        for port in port_urls:
            port = EthernetPort(uri=port, ralph=self.ralph)
            try:
                port.parse()
            except RalphAssetMimatch:
                continue
            self.components['port-' + str(port_index)] = port
            port_index += 1

    def __str__(self):
        ret = super().__str__()
        return ret + '\n\t' + str(self.model)


from abc import ABC, abstractmethod
from typing import List, Dict
import re
import logging

import json
import pyjq
from enum import Enum, auto

from fimutil.ralph.ralph_uri import RalphURI


class RalphAssetType(Enum):
    Node = auto()
    NVMe = auto()
    GPU = auto()
    Ethernet = auto()
    Model = auto()
    Storage = auto()
    DPSwitch = auto()
    Abstract = auto()

    def __str__(self):
        return self.name


class RalphAsset(ABC):
    """
    Abstract Ralph asset - has one or more REST URIs knows
    how to remap JSON responses from those URIs into needed
    fields.
    """
    FIELD_MAP = str()
    REGEX_FIELDS = {}

    def __init__(self, *, uri: str, ralph: RalphURI):
        self.uri = uri
        self.fieldmap = self.FIELD_MAP
        self.regex_fields = self.REGEX_FIELDS
        self.fields = dict()
        self.type = RalphAssetType.Abstract
        self.ralph = ralph
        self.raw_json_obj = None
        self.components = dict()

    def self_populate(self):
        # save JSON object
        self.raw_json_obj = self.ralph.get_json_object(self.uri)
        # massage results - sometimes they are part of the larger query,
        # sometimes node by itself
        if self.raw_json_obj.get('results', None) is not None:
            self.raw_json_obj = self.raw_json_obj['results'][0]

        self.populate_fields_from_obj(json_obj=self.raw_json_obj)
        # populate regex fields
        for k, v in self.regex_fields.items():
            if self.fields[v[0]] is None:
                continue
            matches = re.match(v[1], self.fields[v[0]])
            if matches is not None:
                self.fields[k] = matches.group(1)

    def populate_fields_from_obj(self, *, json_obj):

        self.fields = pyjq.first(self.fieldmap, json_obj)

    def get_fields(self) -> Dict[str, str]:
        return self.fields.copy()

    def parse(self):
        """
        Parse itself and subcomponents
        """
        self.self_populate()

    def __str__(self):
        ret = list()
        ret.append(str(self.type) + ": " + json.dumps(self.fields))
        for n, comp in self.components.items():
            ret.append('\t' + n + " " + str(comp))
        return "\n".join(ret)

    def __repr__(self):
        return self.__str__()


class RalphJSONError(Exception):
    def __init__(self, msg: str):
        super().__init__(f'RalphJSONError: {msg}')


class RalphAssetMimatch(Exception):
    def __init__(self, msg: str):
        super().__init__(f'Ralph asset mismatch: {msg}')


import setuptools
from fimutil import __VERSION__

with open("README.md", "r") as fh:
  long_description = fh.read()

with open("requirements.txt", "r") as fh:
  requirements = fh.read()

setuptools.setup(
  name="fim_utils",
  version=__VERSION__,
  author="Ilya Baldin, Hussamudin Nasir, Xi Yang",
  description="FABRIC Information Model Library Utilities",
  url="https://github.com/fabric-testbed/information-model-utils",
  long_description="FABRIC Information Model Library Utilities",
  long_description_content_type="text/plain",
  packages=setuptools.find_packages(),
  include_package_data=True,
  scripts=['utilities/scan_worker.py', 'utilities/scan_site.py'],
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  python_requires=">=3.9",
  install_requires=requirements,
  setup_requires=requirements,
)



import logging
import re
from fimutil.ralph.asset import RalphAsset, RalphAssetType, RalphAssetMimatch

from fimutil.ralph.ralph_uri import RalphURI


class EthernetPort(RalphAsset):
    """
    This class knows how to parse necessary worker fields in Ralph
    """
    FIELD_MAP = '{MAC: .mac, Description: .model_name, Speed: .speed, Connection: .label}'
    REGEX_FIELDS = {'Peer_port': ['Connection', ".+port ([\\w]+ [0-9/]+) on.+"]}

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)
        self.type = RalphAssetType.Ethernet

    def parse(self):
        super().parse()
        if self.fields['Connection'] is not None and  \
                ('Management' in self.fields['Connection'] or \
                 'Campus' in self.fields['Connection']):
            raise RalphAssetMimatch('This is not a usable port on dataplane switch')


class EthernetCardPort(EthernetPort):
    """
    This is a port on a allocatable card in worker
    """
    FIELD_MAP = '{MAC: .mac, Description: .model_name, Speed: .speed, Connection: .label}'
    REGEX_FIELDS = {'BDF': ['Description', ".+\\(([0-9a-f:.]+)\\).*"],
                    'Peer_port': ['Connection', ".+port ([\\w]+ [0-9/]+) on.+"],
                    'Model': ['Description', ".+\\[([\\w-]+)\\].*"]}

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)
        self.type = RalphAssetType.Ethernet

    def parse(self):
        super().parse()
        if 'data-sw' not in self.fields['Connection']:
            raise RalphAssetMimatch('This is not a usable card')


import re
import logging

from fimutil.ralph.asset import RalphAsset, RalphAssetType, RalphAssetMimatch

from fimutil.ralph.ralph_uri import RalphURI


class NVMeDrive(RalphAsset):
    """
    This class knows how to parse necessary worker fields in Ralph
    """
    FIELD_MAP = '{SN: .serial_number, Description: .model_name}'
    REGEX_FIELDS = {'BDF': ["Description", ".+\\(([0-9a-f:.]+)\\).*"],
                    'Model': ["Description", ".+NVMe ([\\w]+).*"]}

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)
        self.type = RalphAssetType.NVMe

    def parse(self):
        super().parse()
        if 'NVMe' not in self.fields['Description']:
            raise RalphAssetMimatch('This is not an NVMe drive')




import dataclasses

import pyjq
import logging

from fimutil.ralph.asset import RalphAsset, RalphAssetType, RalphJSONError, RalphAssetMimatch
from fimutil.ralph.nvme import NVMeDrive
from fimutil.ralph.ethernetport import EthernetCardPort
from fimutil.ralph.gpu import GPU
from fimutil.ralph.model import WorkerModel
from fimutil.ralph.ralph_uri import RalphURI


class WorkerNode(RalphAsset):
    """
    This class knows how to parse necessary worker fields in Ralph
    """
    FIELD_MAP = '{Name: .hostname, SN: .sn}'

    def __init__(self, *, uri: str, ralph: RalphURI):
        super().__init__(uri=uri, ralph=ralph)
        self.type = RalphAssetType.Node
        self.model = None

    def parse(self):
        super().parse()

        # find model
        model_url = pyjq.one('.model.url', self.raw_json_obj)
        self.model = WorkerModel(uri=model_url, ralph=self.ralph)
        try:
            self.model.parse()
        except RalphAssetMimatch:
            pass

        # find NVMe drives in 'disks' section
        try:
            disk_urls = pyjq.all('.disk[].url', self.raw_json_obj)
        except ValueError:
            logging.warning('Unable to find any disks in node, continuing')
            disk_urls = list()

        disk_index = 1
        for disk in disk_urls:
            drive = NVMeDrive(uri=disk, ralph=self.ralph)
            try:
                drive.parse()
            except RalphAssetMimatch:
                continue
            self.components['nvme-' + str(disk_index)] = drive
            disk_index += 1

        try:
            port_urls = pyjq.all('.ethernet[].url', self.raw_json_obj)
        except ValueError:
            logging.warning('Unable to find any ethernet ports in node, continuing')
            port_urls = list()

        port_index = 1
        for port in port_urls:
            port = EthernetCardPort(uri=port, ralph=self.ralph)
            try:
                port.parse()
            except RalphAssetMimatch:
                continue
            self.components['port-' + str(port_index)] = port
            port_index += 1

        gpus = GPU.find_gpus(self.raw_json_obj)
        gpu_index = 1
        for gpu in gpus:
            self.components['gpu-' + str(gpu_index)] = dataclasses.asdict(gpu)
            gpu_index += 1

    def __str__(self):
        ret = super().__str__()
        return ret + '\n\t' + str(self.model)




